/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <hw/external_connection.h>
#include <libgdbus/dbus-system.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/devices.h"
#include "core/config-parser.h"
#include "core/device-notifier.h"
#include "core/udev.h"
#include "extcon.h"

#define EXTCON_PATH "/sys/class/extcon"
#define STATE_NAME  "STATE"

#define METHOD_SYSPOPUP_SHOW		"show"

#define BUF_MAX 256

static dd_list *extcon_list;

static struct external_connection_device *extcon_dev;

void add_extcon(struct extcon_ops *dev)
{
	DD_LIST_APPEND(extcon_list, dev);
}

void remove_extcon(struct extcon_ops *dev)
{
	DD_LIST_REMOVE(extcon_list, dev);
}

static struct extcon_ops *find_extcon(const char *name)
{
	dd_list *l;
	struct extcon_ops *dev;

	if (!name)
		return NULL;

	DD_LIST_FOREACH(extcon_list, l, dev) {
		if (!strcmp(dev->name, name))
			return dev;
	}

	return NULL;
}

int extcon_get_status(const char *name)
{
	struct extcon_ops *dev;

	if (!name)
		return -EINVAL;

	dev = find_extcon(name);
	if (!dev)
		return -ENOENT;

	if (!dev->enabled)
		return -ENODEV;

	return dev->status;
}

static int extcon_update(const char *name, const char *value)
{
	struct extcon_ops *dev;
	char buf[BUF_MAX];
	int status;
	int ret = 0;

	if (!name || !value)
		return -EINVAL;

	dev = find_extcon(name);
	if (!dev)
		return -EINVAL;

	status = atoi(value);

	/* Do not invoke update func. if it's the same value */
	if (dev->status == status)
		return 0;

	_I("Changed %s device : %d -> %d", name, dev->status, status);

	dev->status = status;

	if (dev->enabled == false && strncmp(name, "USB", strlen("USB")) == 0) {
		if (status > 0) {
			snprintf(buf, BUF_MAX, "usb-client");
			ret = dbus_handle_method_sync_var(DEVICEMANAGER_BUS_NAME,
							DEVICEMANAGER_PATH_POPUP,
							DEVICEMANAGER_INTERFACE_POPUP,
							METHOD_SYSPOPUP_SHOW,
							g_variant_new("(s)", buf));
			if (ret < 0)
				_E("Failed to launch USB restricted popup(%d)", ret);
			return ret;
		} else
			return 0;
	}

	if (dev->update)
		dev->update(status);

	return 0;
}

int extcon_enable_device(const char *name)
{
	struct extcon_ops *dev;
	int ret;

	dev = find_extcon(name);
	if (!dev)
		return -ENODEV;

	if (dev->enabled) {
		_I("extcon(%s) already enabled", name);
		return 0;
	}

	dev->init(NULL);
	dev->enabled = true;

	if (strncmp(name, "USB", strlen("USB")) == 0) {
		ret = extcon_update("USB", "0");
		if (ret != 0)
			_E("fail to disconnect USB");

		ret = extcon_update("USB", "1");
		if (ret != 0)
			_E("fail to connect USB");
	}
	_I("extcon(%s) enabled", name);

	return 0;
}

int extcon_disable_device(const char *name)
{
	struct extcon_ops *dev;
	char buf[BUF_MAX];
	int ret;

	dev = find_extcon(name);
	if (!dev)
		return -ENODEV;

	snprintf(buf, BUF_MAX, "usb-client");
	ret = dbus_handle_method_sync_var(DEVICEMANAGER_BUS_NAME,
					DEVICEMANAGER_PATH_POPUP,
					DEVICEMANAGER_INTERFACE_POPUP,
					METHOD_SYSPOPUP_SHOW,
					g_variant_new("(s)", buf));
	if (ret < 0)
		_E("Failed to launch USB restricted popup(%d)", ret);

	if (!dev->enabled) {
		_I("extcon(%s) already disabled", name);
		return 0;
	}

	dev->exit(NULL);
	dev->enabled = false;

	_I("extcon(%s) disabled", name);

	return 0;
}

static int extcon_parsing_value(const char *value)
{
	char *s, *p;
	char name[NAME_MAX];

	if (!value)
		return -EINVAL;

	s = (char*)value;
	while (s && *s != '\0') {
		p = strchr(s, '=');
		if (!p)
			break;
		memset(name, 0, sizeof(name));
		memcpy(name, s, p-s);
		/* name is env_name and p+1 is env_value */
		extcon_update(name, p+1);
		s = strchr(p, '\n');
		if (!s)
			break;
		s += 1;
	}

	return 0;
}

static void uevent_extcon_handler(struct udev_device *dev)
{
	const char *env_value;
	int ret;

	env_value = udev_device_get_property_value(dev, STATE_NAME);
	if (!env_value)
		return;

	ret = extcon_parsing_value(env_value);
	if (ret < 0)
		_E("fail to parse extcon value : %d", ret);
}

static int extcon_load_uevent(struct parse_result *result, void *user_data)
{
	if (!result)
		return 0;

	if (!result->name || !result->value)
		return 0;

	extcon_update(result->name, result->value);

	return 0;
}

static int get_extcon_init_state(void)
{
	DIR *dir;
	struct dirent *result;
	char node[BUF_MAX];
	int ret;

	dir = opendir(EXTCON_PATH);
	if (!dir) {
		ret = -errno;
		_E("Cannot open dir (%s, errno:%d)", EXTCON_PATH, ret);
		return ret;
	}

	ret = -ENOENT;
	while ((result = readdir(dir))) {
		if (result->d_name[0] == '.')
			continue;
		snprintf(node, sizeof(node), "%s/%s/state",
				EXTCON_PATH, result->d_name);
		_I("checking node (%s)", node);
		if (access(node, F_OK) != 0)
			continue;

		ret = config_parse(node, extcon_load_uevent, NULL);
		if (ret < 0)
			_E("fail to parse %s data : %d", node, ret);
	}

	if (dir)
		closedir(dir);

	return 0;
}

static GVariant * dbus_get_extcon_status(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct extcon_ops *dev;
	char *str;
	int ret;

	g_variant_get(param, "(s)", &str);

	dev = find_extcon(str);
	if (!dev) {
		_E("fail to matched extcon device : %s", str);
		ret = -ENOENT;
		goto error;
	}

	ret = dev->status;
	_D("Extcon device : %s, status : %d", dev->name, dev->status);

error:
	g_free(str);
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_enable_device(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	char *device;
	int ret;

	g_variant_get(param, "(s)", &device);

	ret = extcon_update(device, "1");

	g_free(device);
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_disable_device(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	char *device;
	int ret;

	g_variant_get(param, "(s)", &device);

	ret = extcon_update(device, "0");

	g_free(device);
	return g_variant_new("(i)", ret);
}

static struct uevent_handler uh = {
	.subsystem = EXTCON_SUBSYSTEM,
	.uevent_func = uevent_extcon_handler,
};

static const dbus_method_s dbus_methods[] = {
	{ "GetStatus", "s",  "i", dbus_get_extcon_status },
	{ "enable",    "s", "i", dbus_enable_device },  /* for devicectl */
	{ "disable",   "s", "i", dbus_disable_device }, /* for devicectl */
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_EXTCON,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};


/* used by HAL */
static void extcon_changed(struct connection_info *info, void *data)
{
	if (!info)
		return;

	if (!info->name || !info->state)
		return;

	/* call to update */
	extcon_update(info->name, info->state);
}

static int extcon_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (extcon_dev)
		return 0;

	ret = hw_get_info(EXTERNAL_CONNECTION_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret == 0) {
		if (!info->open) {
			_E("Failed to open extcon device; open(NULL)");
			return -ENODEV;
		}

		ret = info->open(info, NULL, (struct hw_common **)&extcon_dev);
		if (ret < 0) {
			_E("Failed to get extcon device structure (%d)", ret);
			return ret;
		}

		_I("extcon device structure load success");
		return 0;
	}

	/**
	 * find extcon class.
	 * if there is no extcon class,
	 * deviced does not control extcon devices.
	 */
	if (access(EXTCON_PATH, R_OK) != 0) {
		_E("there is no extcon class");
		return -ENODEV;
	}

	return 0;
}

static void add_extcon_event_handler(void)
{
	int ret;

	if (extcon_dev) { /* HAL is used */
		if (extcon_dev->register_changed_event)
			extcon_dev->register_changed_event(extcon_changed, NULL);

		if (extcon_dev->get_current_state)
			extcon_dev->get_current_state(extcon_changed, NULL);

	} else {
		/* register extcon uevent */
		ret = register_kernel_uevent_control(&uh);
		if (ret < 0)
			_E("fail to register extcon uevent : %d", ret);

		/* load the initialize value by accessing the node directly */
		ret = get_extcon_init_state();
		if (ret < 0)
			_E("fail to init extcon nodes : %d", ret);
	}
}

static void remove_extcon_event_handler(void)
{
	int ret;

	if (extcon_dev) {
		if (extcon_dev->unregister_changed_event)
			extcon_dev->unregister_changed_event(extcon_changed);
	} else {
		/* unreigster extcon uevent */
		ret = unregister_kernel_uevent_control(&uh);
		if (ret < 0)
			_E("fail to unregister extcon uevent : %d", ret);
	}
}

static int event_handler_state_changed(void *data)
{
	static device_notifier_state_e old = DEVICE_NOTIFIER_STATE_STOP;
	device_notifier_state_e state = *(device_notifier_state_e *)data;

	if (old == state)
		return 0;

	old = state;

	if (state == DEVICE_NOTIFIER_STATE_START)
		add_extcon_event_handler();
	else if (state == DEVICE_NOTIFIER_STATE_STOP)
		remove_extcon_event_handler();

	return 0;
}

static void extcon_init(void *data)
{
	int ret;
	dd_list *l;
	struct extcon_ops *dev;
	device_notifier_state_e state = DEVICE_NOTIFIER_STATE_START;

	if (!extcon_list)
		return;

	/* initialize extcon devices */
	DD_LIST_FOREACH(extcon_list, l, dev) {
		_I("[extcon] init (%s)", dev->name);
		if (dev->init)
			dev->init(data);
		dev->enabled = true;
	}

	event_handler_state_changed((void *)&state);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_EXTCON, &dbus_interface);
	if (ret < 0)
		_E("fail to init dbus method(%d)", ret);


	register_notifier(DEVICE_NOTIFIER_EVENT_HANDLER, event_handler_state_changed);
}

static void extcon_exit(void *data)
{
	dd_list *l;
	struct extcon_ops *dev;
	device_notifier_state_e state = DEVICE_NOTIFIER_STATE_STOP;

	unregister_notifier(DEVICE_NOTIFIER_EVENT_HANDLER, event_handler_state_changed);

	event_handler_state_changed((void *)&state);

	/* deinitialize extcon devices */
	DD_LIST_FOREACH(extcon_list, l, dev) {
		_I("[extcon] deinit (%s)", dev->name);
		if (dev->exit)
			dev->exit(data);
		dev->enabled = false;
	}
}

static const struct device_ops extcon_device_ops = {
	.name   = "extcon",
	.probe  = extcon_probe,
	.init   = extcon_init,
	.exit   = extcon_exit,
};

DEVICE_OPS_REGISTER(&extcon_device_ops)
