/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <device-node.h>
#include <vconf.h>
#include <fcntl.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/devices.h"
#include "extcon_count.h"

#define BUF_MAX 256

// TODO : storage location should be configurable by configuration file
static struct extcon_device {
	const enum extcon_type type;
	const char *str;
	int fd;
	unsigned long count;
} extcon_devices[] = {
	{ EXTCON_TA,           "/opt/share/batt_cable_count",   0, 0},
	{ EXTCON_EARJACK,      "/opt/share/earjack_count",      0, 0},
	{ EXTCON_BATTERY_FULL, "/opt/share/battery_full_count", 0, 0},
};

int extcon_get_count(int index)
{
	int fd;
	int r;
	int ret = 0;
	char buf[BUF_MAX];

	if (extcon_devices[index].fd) {
		close(extcon_devices[index].fd);
		extcon_devices[index].fd = 0;
	}
	fd = open(extcon_devices[index].str, O_RDWR);
	if (fd < 0)
		return -ENOENT;

	r = read(fd, buf, BUF_MAX);
	if (r < 0 || r >= BUF_MAX) {
		close(fd);
		ret = -EIO;
		return ret;
	}

	buf[r] = '\0';
	extcon_devices[index].fd = fd;
	extcon_devices[index].count = strtoul(buf, NULL, 10);
	_I("get extcon(%d:%x) count %lu",
		index, extcon_devices[index].fd, extcon_devices[index].count);

	return ret;
}

int extcon_set_count(int index, unsigned long count)
{
	int r;
	int ret = 0;
	char buf[BUF_MAX];

	extcon_devices[index].count += count;

	if (extcon_devices[index].fd < 0) {
		_E("cannot open file(%s)", extcon_devices[index].str);
		return -ENOENT;
	}
	lseek(extcon_devices[index].fd, 0, SEEK_SET);
	_I("ext(%d) count %lu", index, extcon_devices[index].count);
	snprintf(buf, sizeof(buf), "%lu", extcon_devices[index].count);

	r = write(extcon_devices[index].fd, buf, strlen(buf));
	if (r < 0)
		ret = -EIO;
	return ret;
}

static int extcon_create_count(int index)
{
	int fd;
	int r;
	int ret = 0;
	char buf[BUF_MAX];
	fd = open(extcon_devices[index].str, O_RDWR | O_CREAT, 0644);
	if (fd < 0) {
		_E("cannot open file(%s)", extcon_devices[index].str);
		return -ENOENT;
	}
	snprintf(buf, sizeof(buf), "%lu", extcon_devices[index].count);
	r = write(fd, buf, strlen(buf));
	if (r < 0)
		ret = -EIO;

	if (ret != 0) {
		close(fd);
		_E("cannot write file(%s)", extcon_devices[index].str);
		return ret;
	}
	extcon_devices[index].fd = fd;
	_I("create extcon(%d:%x) %s",
		index, extcon_devices[index].fd, extcon_devices[index].str);
	return ret;
}

static void extcon_count_init(void *data)
{
	int i, ret;

	for (i = 0; i < ARRAY_SIZE(extcon_devices); i++) {
		if (extcon_get_count(i) >= 0)
			continue;
		ret = extcon_create_count(i);
		if (ret < 0)
			break;
	}
}

static const struct device_ops extcon_count_ops = {
	.name   = "extcon_count",
	.init   = extcon_count_init,
};

DEVICE_OPS_REGISTER(&extcon_count_ops)
